# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2O10 Nexedi SA and Contributors. All Rights Reserved.
#                    Jean-Paul Smets-Solanes <jp@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import types

from Products.CMFCore.utils import getToolByName
from Products.ERP5Type.TransactionalVariable import getTransactionalVariable

class ExplanationCache:
  """ExplanationCache provides a central access to 
  all parameters and values which are needed to process 
  an explanation. It is based on the idea that a value is calculated
  once and once only, as a way to accelerate performance of algorithms
  related to an explanation.
  """

  def __init__(self, explanation):
    """Explanation cache keeps a handful list of caches
    to accelerate performance of business path browsing and
    business process algorithms
    """
    # Define share properties
    self.explanation = explanation
    self.portal_catalog = getToolByName(explanation, 'portal_catalog')
    self.simulation_movement_cache = {} # Simulation Movement Cache
    self.explanation_uid_cache = []
    self.explanation_path_pattern_cache = []
    self.closure_cache = []
    self.union_cache = None

  def _getDeliveryMovementList(self):
    """Returns self is explanation is a delivery line
    of the list of explanation delivery lines if explanation
    is a delivery
    """
    explanation = self.explanation
    if explanation.isDelivery():
      # Gather all movements of the delivery
      delivery_movement_list = explanation.getMovementList()
    else:
      # Only consider a single movement - XXX-JPS is this OK when we have lines in lines ?
      delivery_movement_list = [explanation]
    return delivery_movement_list

  def getRootExplanationUidList(self):
    """Return the list of explanation_uid values involved
    in the context of the explanation. This will be useful later
    in order to accelerate searches in the catalog.
    """
    # Return cache if defined
    if self.explanation_uid_cache:
      return self.explanation_uid_cache
    result = set()
    # For each delivery movement
    for movement in self._getDeliveryMovementList():
      # For each simulation movement
      for simulation_movement in movement.getDeliveryRelatedValueList():
        result.add(simulation_movement.getExplanationUid()) # XXX-JPS use new API later
    # Return result
    self.explanation_uid_cache = result
    return result

  def getSimulationPathPatternList(self):
    """Return the list of root path of simulation tree which are 
    involved in the context of the explanation. This will be useful later
    in order to accelerate searches in the catalog.

    XXX-JPS: QUESTION: should we consider only patterns starting from
    the movement, or from the root delivery line related movement ?
    In one case, we must provided appropriate explanation for everything
    to work. In the other case, we can take any delivery line any where
    a explanation.
    """
    # Return cache if defined
    if self.explanation_path_pattern_cache:
      return self.explanation_path_pattern_cache

    # Helper method to update path_dict with
    # each key which forms the pay of the simulation_movement
    path_dict = {}
    def updatePathDict(simulation_movement):
      local_path_dict = path_dict
      container_path = simulation_movement.getParentValue().getPhysicalPath()
      simulation_movement_id = simulation_movement.getId()
      insert_movement = True
      for path_id in container_path:
        if local_path_dict.get(path_id, None) is None:
          local_path_dict[path_id] = dict()
        local_path_dict = local_path_dict[path_id]
        if type(local_path_dict) is not types.DictType:
          # A movement was already inserted
          insert_movement = False
          break
      if insert_movement:
        local_path_dict[simulation_movement_id] = simulation_movement

    # For each delivery movement
    for movement in self._getDeliveryMovementList():
      # For each simulation movement
      for simulation_movement in movement.getDeliveryRelatedValueList():
        updatePathDict(simulation_movement)
    
    # Build result by browsing path_dict and
    # assembling path '/erp5/portal_simulation/1/34/23/43%'
    result = []
    def browsePathDict(prefix, local_path_dict):
      for key, value in local_path_dict.items():
        if type(value) is not types.DictType:
          # We have a real root
          result.append('%s/%s%' % (prefix, key))
        else:
          browsePathDict('%s/%s' % (prefix, key), value) # Recursing with string append is slow XXX-JPS

    browsePathDict('/', path_dict)
    self.explanation_path_pattern_cache = result
    return result

  def getBusinessPathRelatedSimulationMovementValueList(self, business_path):
    """Returns the list of simulation movements caused by a business_path
    in the context the our explanation.
    """
    return self.getSimulationMovementValueList(causality_uid=business_path.getUid())
    
  def getSimulationMovementValueList(self, **kw):
    """Search Simulation Movements related to our explanation.
    Cache result so that the second time we saarch for the same
    list we need not involve the catalog again.
    """
    kw_tuple = tuple(kw.items()) # We hope that no sorting is needed
    if kw.get('path', None) is None:
      kw['path'] = self.getSimulationPathPatternList(), # XXX-JPS Explicit Query is better
    if kw.get('explanation_uid', None) is None:
      kw['explanation_uid'] = self.getRootExplanationUidList()
    if self.simulation_movement_list.get(kw_tuple, None) is None:
      self.simulation_movement_cache[kw_tuple] = \
           self.portal_catalog(portal_type="Simulation Movement",
                               explanation_uid=explanation_uid,
                               path=path,
                               **kw)
    return self.simulation_movement_cache[kw_tuple]

  def getBusinessPathValueList(self, **kw):
    """Find all business path which are related to the simulation 
    trees defined by the explanation.
    """
    business_type_list = self.getPortalBusinessPathTypeList()
    simulation_movement_list = self.getSimulationMovementValueList()
    simulation_movement_uid_list = map(lambda x:x.uid, simulation_movement_list) 
    # We could use related keys instead of 2 queries
    business_path_list = self.portal_catalog(
                      portal_type=business_type_list,
                      causality_related_uid=simulation_movement_uid_list,
                      **kw)
    return business_path_list

  def geBusinessPathClosure(business_process, business_path):
    """Creates a Business Process by filtering out all Business Path
    in 'business_process' which are not related to a simulation movement
    which is either or parent or a child of explanation simulations movements
    caused by 'business_path'
    """
    # Try to return cached value first
    new_business_process = self.closure_cache.get(business_path, None)
    if new_business_process is not None:
      return new_business_process

    # Build a list of path patterns which apply to current business_path
    path_list = self.getSimulationPathPatternList()
    path_list = map(lambda x:x[0:-1], path_list) # Remove trailing %
    path_set = set()
    for simulation_movement in business_path.\
             _getExplanationRelatedSimulationMovementValueList(explanation):
      simulation_path = simulation_movement.getPath()
      for path in path_list:
        if simulation_path.startswith(path):
          path_set.add(path) # Only keep a path pattern which matches current simulation movement

    # Lookup in cache based on path_tuple
    path_tuple = tuple(path_set) # XXX-JPS is the order guaranteed here ?
    new_business_process = self.closure_cache.get(path_tuple, None)
    if new_business_process is not None:
      self.closure_cache[business_path] = new_business_process
      return new_business_process

    # Build a new closure business process
    def hasMatchingMovement(business_path):
      return len(self.getSimulationMovementValueList(path=path_set, 
                                  causality_uid=business_path.getUid()))      

    new_business_process = business_process.getParentValue().newContent(temp_object=True) # XXX-JPS is this really OK with union business processes
    i = 0
    for business_path in business_process.getBusinessPathValueList():
      if hasMatchingMovement(business_path):
        i += 1
        id = 'closure_path_%s' % i
        new_business_process._setOb(id, business_path.asContext(id=id))

    self.closure_cache[business_path] = new_business_process
    self.closure_cache[path_tuple] = new_business_process
    return new_business_process

  def getUnionBusinessProcess(self):
    """Return a Business Process made of all Business Path
    which are the cause of Simulation Movements in the simulation
    trees related to explanation.
    """
    # Try to return cached value first
    new_business_process = self.union_cache
    if new_business_process is not None:
      return new_business_process

    # Build Union Business Process
    from Products.ERP5Type.Document import newTempBusinessProcess
    new_business_process = newTempBusinessProcess(self.explanation, 'union_business_process')
    i = 0
    for business_path in self.getBusinessPathValueList():
      i += 1
      id = 'union_path_%s' % i
      new_business_process._setOb(id, business_path.asContext(id=id))

    # Keep it in cache and return
    self.union_cache = new_business_process
    return new_business_process


def _getExplanationCache(explanation):
  # Return cached value if any
  tv = getTransactionalVariable(explanation)
  if tv.get('explanation_cache', None) is None:
    tv['explanation_cache'] =  ExplanationCache(explanation)
  return tv.get('explanation_cache')

def _getBusinessPathClosure(business_process, explanation, business_path):
  """Returns a closure Business Process for given 
  business_path and explanation. This Business Process
  contains only those Business Path which are related to business_path
  in the context of explanation.
  """
  explanation_cache = _getExplanationCache(explanation)
  return explanation_cache.getBusinessPathClosure(business_process, business_path)

def _getUnionBusinessProcess(explanation):
  """Build a Business Process by taking the union of Business Path
  which are involved in the simulation trees related to explanation
  """
  explanation_cache = _getExplanationCache(explanation)
  return explanation_cache.getUnionBusinessProcess()